// Выбор текущего адреса
assign address = src? {segment, 4'h0} + ea : {cs,4'h0} + ip;

// Инициализация выходов
initial begin o_data = 8'b0; we = 1'b0; end

// -----------------------------------------------

// Знакорасширенный i_data
wire [15:0] signex = {{8{i_data[7]}}, i_data[7:0]};

// -----------------------------------------------

// Регистры общего назначения
reg [15:0]  ax = 16'h0540;
reg [15:0]  bx = 16'h1202;
reg [15:0]  cx = 16'h5678;
reg [15:0]  dx = 16'h7EE1;
reg [15:0]  sp = 16'h3411;
reg [15:0]  bp = 16'hFAFE;
reg [15:0]  si = 16'hBABE;
reg [15:0]  di = 16'hB0BA;

// Сегментные регистры
reg [15:0]  es = 16'h0000;
reg [15:0]  cs = 16'hF000;
reg [15:0]  ss = 16'h0005;
reg [15:0]  ds = 16'h0000;

// Флаги                ODIT SZ A  P C
reg [11:0]  flags   = 12'b0000_0000_0000;
reg [15:0]  ip      = 16'h0000; // FFF0

// Состояние процессора
// -----------------------------------------------
reg [ 5:0]  t           = 1'b0;
reg [ 2:0]  m           = 1'b0;
reg [ 5:0]  tnext       = 1'b0;
reg [ 7:0]  opcode      = 1'b0;
reg [15:0]  segment     = 16'h0000;
reg [15:0]  ea          = 16'h0000;
reg         segover     = 1'b0;
reg [ 1:0]  rep         = 2'b00;
reg         dir         = 1'b0;
reg         src         = 1'b0;
reg         size        = 1'b0;
reg [ 2:0]  alu         = 1'b0;
reg [ 7:0]  modrm       = 8'h00;
reg [15:0]  op1         = 16'h0000;
reg [15:0]  op2         = 16'h0000;
reg [15:0]  wb          = 16'h0000;
reg [ 2:0]  regn        = 3'h0;

// Выборка регистра regv из regn
// -----------------------------------------------
wire [15:0] regv =
    regn == 3'h0 ? (size ? ax : ax[ 7:0]) :
    regn == 3'h1 ? (size ? cx : cx[ 7:0]) :
    regn == 3'h2 ? (size ? dx : dx[ 7:0]) :
    regn == 3'h3 ? (size ? bx : bx[ 7:0]) :
    regn == 3'h4 ? (size ? sp : ax[15:8]) :
    regn == 3'h5 ? (size ? bp : cx[15:8]) :
    regn == 3'h6 ? (size ? si : dx[15:8]) :
                   (size ? di : bx[15:8]);

// Вычисление условий
wire [7:0] branches = {

    /*7*/ (flags[SF] ^ flags[OF]) | flags[ZF],
    /*6*/ (flags[SF] ^ flags[OF]),
    /*5*/  flags[PF],
    /*4*/  flags[SF],
    /*3*/  flags[CF] | flags[ZF],
    /*2*/  flags[ZF],
    /*1*/  flags[CF],
    /*0*/  flags[OF]
};

// -----------------------------------------------
localparam
    init        = 1'b0,     // Стадия подготовки инструкции
    fetch       = 1'b1,     // Считывание опкода и префиксов
    exec        = 2'h2,     // Исполнение инструкции
    // Для считывания ModRM
    tmodrm      = 2'h3,     // Считывание байта modrm
    tmodrm1     = 3'h4,     // +d8
    tmodrm2     = 3'h5,     // +d16 low
    tmodrmh     = 3'h6,     // +d16 hi
    tmodrmf     = 4'h8,     // Чтение операндов
    tmodrmf2    = 4'h9,     // Чтение операндов (high)
    wback       = 4'hA,     // Запись [wb,size,dir,modrm]
    wback2      = 4'hB,
    wback3      = 4'hC;
